package websocket_utils

import (
	"context"
	"fmt"
	"github.com/aws/aws-lambda-go/events"
	"github.com/aws/aws-sdk-go-v2/aws"
	"github.com/aws/aws-sdk-go-v2/config"
	"github.com/aws/aws-sdk-go-v2/service/apigatewaymanagementapi"
	"gitlab.bob.co.za/bob-public-utils/bobgroup-go-utils/logs"
	"gitlab.bob.co.za/bob-public-utils/bobgroup-go-utils/utils"
	"os"
)

var (
	sessions = map[string]*APIGateWayClientWithHelpers{}
)

type APIGateWayClientWithHelpers struct {
	APIGatewayClient *apigatewaymanagementapi.Client
}

func GetClient(req *events.APIGatewayWebsocketProxyRequest, region ...string) *APIGateWayClientWithHelpers {
	if req == nil {
		logs.ErrorMsg("APIGatewayWebsocketProxyRequest is nil")
		return nil
	}

	s3Region := os.Getenv("AWS_REGION")

	// Set custom region
	if region != nil && len(region) > 0 {
		s3Region = region[0]
	}

	// Check if session exists for region, if it does return it
	if apiGatewaySession, ok := sessions[s3Region]; ok {
		return apiGatewaySession
	}

	// Setup session
	cfg, err := config.LoadDefaultConfig(context.TODO(),
		config.WithRegion(s3Region),
	)
	if err != nil {
		return nil
	}

	apiGatewaySession := NewClient(req, cfg)
	sessions[s3Region] = apiGatewaySession
	return apiGatewaySession
}

func NewClient(req *events.APIGatewayWebsocketProxyRequest, config aws.Config) *APIGateWayClientWithHelpers {
	return &APIGateWayClientWithHelpers{
		APIGatewayClient: apigatewaymanagementapi.NewFromConfig(config, func(o *apigatewaymanagementapi.Options) {
			o.BaseEndpoint = utils.ValueToPointer(fmt.Sprintf("https://%s/%s", req.RequestContext.DomainName, req.RequestContext.Stage))
		}),
	}
}

func (s APIGateWayClientWithHelpers) PostToConnectionIDs(data []byte, connectionIDs []string) error {
	for _, connectionID := range connectionIDs {
		_, err := s.APIGatewayClient.PostToConnection(context.TODO(), &apigatewaymanagementapi.PostToConnectionInput{
			ConnectionId: &connectionID,
			Data:         data,
		})
		if err != nil {
			return err
		}
	}

	return nil
}
